/* 
   Program generate triangulations by adding faces in lexigraphic order.
*/

/* version 0.15 */

#define USAGE \
" lextri [-v] [-h] [-i] [-o o] [-e e] [-d d] [-g g] [-G G] [-r res -m mod] nv"

#define HELPTEXT \
" lextri : generate triangulations of surfaces with nv vertices.\n\
\n\
 help switches:\n\
   -h  show this text\n\
   -v  verbose output\n\
\n\
 parameter switches:\n\
   -i          if present only irreducible triangulations are generated\n\
   -o o        0: only nonorientable, 1: only orientable, otherwise: both\n\
   -e e        only equivelar with degree e are generated\n\
   -d d        minimum vertex degree\n\
   -g g        minimum Euler genus\n\
   -G G        maximum Euler genus\n\
   -r res      res for splitting\n\
   -m mod      mod for splitting\n\
\n\
   nv          number of vertices in generated triangulations.\n"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#define CPUTIME 1          /* Whether to measure the cpu time or not */

#if CPUTIME
#include <sys/times.h>
#ifndef CLK_TCK
#include <time.h>
#endif
#endif

#define FALSE 0
#define TRUE 1

#define MAXN 110
#define MAXF 500
#define MAXG 50

/* BIG macros taken from plantri. */
/* The program is so fast that the count of output graphs can quickly
   overflow a 32-bit integer.  Therefore, we use two long values
   for each count, with a ratio of 10^9 between them.  The macro
   ADDBIG adds a small number to one of these big numbers.  
   BIGTODOUBLE converts a big number to a double (approximately).
   SUMBIGS adds a second big number into a first big number.
   SUBBIGS subtracts a second big number from a first big number.
   PRINTBIG prints a big number in decimal.
   ZEROBIG sets the value of a big number to 0.
   ISZEROBIG tests if the value is 0.
   SETBIG sets a big number to a value at most 10^9-1.
   ISEQBIG tests if two big numbers are equal.
*/

typedef struct
{
    long hi,lo;
} bigint;

#define ZEROBIG(big) big.hi = big.lo = 0L
#define ISZEROBIG(big) (big.lo == 0 && big.hi == 0)
#define SETBIG(big,value) {big.hi = 0L; big.lo = (value);}
#define ADDBIG(big,extra) if ((big.lo += (extra)) >= 1000000000L) \
    { ++big.hi; big.lo -= 1000000000L;}
#define PRINTBIG(file,big) if (big.hi == 0) \
 fprintf(file,"%ld",big.lo); else fprintf(file,"%ld%09ld",big.hi,big.lo)
#define BIGTODOUBLE(big) (big.hi * 1000000000.0 + big.lo)
#define SUMBIGS(big1,big2) {if ((big1.lo += big2.lo) >= 1000000000L) \
    {big1.lo -= 1000000000L; big1.hi += big2.hi + 1L;} \
    else big1.hi += big2.hi;}
#define SUBBIGS(big1,big2) {if ((big1.lo -= big2.lo) < 0L) \
    {big1.lo += 1000000000L; big1.hi -= big2.hi + 1L;} \
    else big1.hi -= big2.hi;}
/* Note: SUBBIGS must not allow the value to go negative.
   SUMBIGS and SUBBIGS both permit big1 and big2 to be the same bigint. */
#define ISEQBIG(big1,big2) (big1.lo == big2.lo && big1.hi == big2.hi)

static int verbose;
static int only_irreducible; /* flag if only irreducible triangulations are to
				be generated */
static int nv;     /* number of vertices in final triangulation */
static int ne;     /* number of edges in triangulation */
static int nf;     /* number of faces in triangulation */
static int minnv;  /* minimum number of vertices in a triangulation */
static int maxnv;  /* maximum number of vertices in a triangulation */
static int minnf;  /* minimum number of faces in a triangulation */
static int maxnf;  /* maximum number of faces in a triangulation */
static int ming;   /* minimum Euler genus of a triangulation */
static int maxg;   /* maximum Euler genus of a triangulation */
static int mindegree; /* minimum degree of vertex of a triangulation */
static int maxdegree; /* maximum degree of vertex of a triangulation */
static int edge[MAXN][MAXN]; /* edge[i][j] and edge[j][i] are the vertices of 
				the two faces sharing the edge (i,j).  
				-1 for no corresponding face. */
static int degree[MAXN]; /* edge degree of vertices */
static int fdegree[MAXN]; /* face degree of vertices, number of faces with 
			     that vertex */
static int complete[MAXN]; /* link of vertex is a triangulation of a circle */
static int completeness[MAXN]; /* 0 if not complete, 1 if newly complete, or 
				  2 if formerly complete */
static int match[MAXN][2*MAXN][MAXN]; /* lk(v0) is equivalent to lk(0),
				      i is an index of the automorphisms of 
				      this equivalence,
				      v is mapped to match[v0][i][v] by the 
				      i-th automorphism */
static int nmatch[MAXN];   /* number of automorphisms above */
static int list[MAXF][3];  /* list of the vertices in each ordered face */
static bigint ngenerated[MAXG][2]; /* number of triangulations generated by 
				      Euler genus and orientability */
static int nt_match;        /* number of triangles matched */
static int v_to_label[MAXN];/* relabeling of vertices */
static int label_to_v[MAXN];/* vertex with label*/
static int next_label;      /* next unused label */
static int debug_count;
static int res,mod;        /* res/mod from command line (default 0/1) */
static char res_text[10];  /* text string for res with padded zeros */
static int splitlevel,
           splitcount;     /* used for res/mod splitting */
static int orientability;  /* orientability of output surfaces, 
			      0 for only nonorientable, 1 for orientable, 
			      otherwise both */
static int equivelar;      /* if > 0, degree for all vertices */
static FILE *outfile_alpha[MAXN][MAXG][2];
static FILE *outfile_lex[MAXN][MAXG][2];

#define MAX(x,y) ((x)<(y) ? (y) : (x))
#define MIN(x,y) ((x)>(y) ? (y) : (x))

#ifdef SPLITTEST
static int splitcases;
#endif

void next_face();

void error_exit(int errornum)
{
  fprintf(stderr,"Error number %d, exiting\n",errornum);
  exit(1);
}

void write_alpha(FILE *output)
{
  /* write out surface ala plantri format */

  int v0,v1,v2,v3,i;

  fprintf(output,"%d ",nv);
  for (v0=0; v0<nv; v0++) {
    for (v1=0; edge[v0][v1] == -1; v1++) {}
    v2 = edge[v0][v1];
    for (i=0; i<degree[v0]; i++) {
      fprintf(output,"%c",v1+'a');
      v3 = v1;
      v1 = v2;
      if (edge[v0][v1] == v3)
	v2 = edge[v1][v0];
      else
	v2 = edge[v0][v1];
    }
    if (v0 < nv-1)
      fprintf(output,",");
    else
      fprintf(output,"\n");
  }
}

void write_lex(FILE *output)
{
  /* write out surface in sortable format.  one line of text for each surface.
     one character for each vertex of each face. */

  int iface;

  for (iface=0; iface<nf; iface++)
    fprintf(output,"%c%c%c",
	    list[iface][0]+'a',list[iface][1]+'a',list[iface][2]+'a');
  fprintf(output,"\n");
}

void dump_it()
{
  int ilist;
  int v0,v1;

  fprintf(stderr,"nv = %d, ne = %d, nf = %d\n",nv,ne,nf);

  fprintf(stderr,"list of faces:\n");
  for (ilist=0; ilist<nf; ilist++)
    fprintf(stderr,"%2d %2d %2d \n",
	    list[ilist][0],list[ilist][1],list[ilist][2]);

  fprintf(stderr,"edge matrix:\n");
  fprintf(stderr,"   |");
  for (v1=0; v1<nv; v1++) 
    fprintf(stderr,"%2d ",v1);
  fprintf(stderr,"\n");
  fprintf(stderr,"----");
  for (v1=0; v1<nv; v1++) 
    fprintf(stderr,"---");
  fprintf(stderr,"\n");
  for (v0=0; v0<nv; v0++) {
    fprintf(stderr,"%2d |",v0);
    for (v1=0; v1<nv; v1++) 
      if (edge[v0][v1] != -1)
	fprintf(stderr,"%2d ",edge[v0][v1]);
      else
	fprintf(stderr,"   ");
    fprintf(stderr,"\n");
  }

  fprintf(stderr,"edge degree:");
  for (v1=0; v1<nv; v1++) 
    fprintf(stderr,"%2d ",degree[v1]);
  fprintf(stderr,"\n");

  fprintf(stderr,"face degree:");
  for (v1=0; v1<nv; v1++) 
    fprintf(stderr,"%2d ",fdegree[v1]);
  fprintf(stderr,"\n");

}

void add_face(int v0, int v1, int v2)
{
  if (0 > v0)
    error_exit(5);

  if (v0 >= v1)
    error_exit(6);

  if (v1 >= v2)
    error_exit(7);

  if (v2 > nv)
    error_exit(8);

  if (edge[v0][v1] != -1 &&
      edge[v1][v0] != -1)
    error_exit(1);
  if (edge[v1][v2] != -1 &&
      edge[v2][v1] != -1)
    error_exit(2);
  if (edge[v2][v0] != -1 &&
      edge[v0][v2] != -1)
    error_exit(3);

  if (nf >= MAXF)
    error_exit(4);
  list[nf][0] = v0;
  list[nf][1] = v1;
  list[nf][2] = v2;
  nf++;

  if (v2 == nv)
    nv++;

  if (edge[v0][v1] == -1) {
    edge[v0][v1] = v2;
    if (edge[v1][v0] == -1) {
      degree[v0]++;
      degree[v1]++;
      ne++;
    }
  }
  else
    edge[v1][v0] = v2;
 
  if (edge[v1][v2] == -1) {
    edge[v1][v2] = v0;
    if (edge[v2][v1] == -1) {
      degree[v1]++;
      degree[v2]++;
      ne++;
    }
  }
  else
    edge[v2][v1] = v0;
 
  if (edge[v2][v0] == -1) {
    edge[v2][v0] = v1;
    if (edge[v0][v2] == -1) {
      degree[v2]++;
      degree[v0]++;
      ne++;
    }
  }
  else
    edge[v0][v2] = v1;

  fdegree[v0]++;
  fdegree[v1]++;
  fdegree[v2]++;

  /* update complete */
  
  if (degree[v0] == fdegree[v0]) {
    complete[v0] = TRUE;
    completeness[v0] = 1;
  }
  if (degree[v1] == fdegree[v1]) {
    complete[v1] = TRUE;
    completeness[v1] = 1;
  }
  if (degree[v2] == fdegree[v2]) {
    complete[v2] = TRUE;
    completeness[v2] = 1;
  }
}

void remove_face()
{
  int v0,v1,v2;

  nf--;
  v0 = list[nf][0];
  v1 = list[nf][1];
  v2 = list[nf][2];

  complete[v0] = FALSE;
  complete[v1] = FALSE;
  complete[v2] = FALSE;
  completeness[v0] = 0;
  completeness[v1] = 0;
  completeness[v2] = 0;

  if (edge[v0][v1] == v2) {
    edge[v0][v1] = -1;
    if (edge[v1][v0] == -1) {
      degree[v0]--;
      degree[v1]--;
      ne--;
    }
  }
  else {
    edge[v1][v0] = -1;
    if (edge[v0][v1] == -1) {
      degree[v0]--;
      degree[v1]--;
      ne--;
    }
  }
    
  if (edge[v1][v2] == v0) {
    edge[v1][v2] = -1;
    if (edge[v2][v1] == -1) {
      degree[v1]--;
      degree[v2]--;
      ne--;
    }
  }
  else {
    edge[v2][v1] = -1;
    if (edge[v1][v2] == -1) {
      degree[v1]--;
      degree[v2]--;
      ne--;
    }
  }
    
  if (edge[v2][v0] == v1) {
    edge[v2][v0] = -1;
    if (edge[v0][v2] == -1) {
      degree[v2]--;
      degree[v0]--;
      ne--;
    }
  }
  else {
    edge[v0][v2] = -1;
    if (edge[v2][v0] == -1) {
      degree[v2]--;
      degree[v0]--;
      ne--;
    }
  }

  fdegree[v0]--;
  fdegree[v1]--;
  fdegree[v2]--;
  if (fdegree[v2] == 0)
    nv--;
}

void initialize()
{
  /* one time initialization */

  int orient,g;

#ifdef SPLITTEST
  splitcases = 0;
#endif

  debug_count = 0;

  for (orient=0; orient<2; orient++)
    for (g=0; g<MAXG; g++)
      ZEROBIG(ngenerated[g][orient]);
}

void initialize_first_vertex(int degree0)

{

  int i,j;

  /* clear data*/

  for (i=0; i<maxnv; i++)
    for (j=0; j<maxnv; j++)
      edge[i][j] = -1;
  for (i=0; i<maxnv; i++) {
    degree[i] = 0;
    fdegree[i] = 0;
    v_to_label[i] = -1;
    complete[i] = FALSE;
  }

  nv = 2;
  ne = 0;
  nf = 0;

  /* add faces around first vertex */

  add_face(0,1,2);
  for (i=3; i<=degree0; i++)
    add_face(0,i-2,i);
  add_face(0,degree0-1,degree0);

} 

int orientable()
{
  /* check if surface is orientable. */
  
  /* only the strongly connected component containing the first face is 
     checked.
     false positive is possible. */
  
  int orient[MAXN][MAXN];  
                   /* orient[i][j] is orientation of face (i,j,edge[i][j]) . 
		      +1 if clockwise,
		      -1 if counter-clockwise,
		      0 if don't know yet. */
  int iface;         /* face index */
  int v0,v1,v2;
  int face_orient[3];
  int up,down,new_orient;
  int ofound[MAXF]; /* orientation of face has been determined */
  int nfound;       /* number of faces for which orientation has been 
		       determined */
  int more_found;
  
  for (v0=0; v0<nv; v0++)
    for (v1=0; v1<nv; v1++)
      orient[v0][v1] = 0;
  for (iface=0; iface<nf; iface++)
    ofound[iface] = FALSE;
  nfound = 0;
  
  if (edge[list[0][0]][list[0][1]] == list[0][2])
    orient[list[0][0]][list[0][1]] = +1;
  else
    orient[list[0][1]][list[0][0]] = -1;
  
  if (edge[list[0][1]][list[0][2]] == list[0][0])
    orient[list[0][1]][list[0][2]] = +1;
  else
    orient[list[0][2]][list[0][1]] = -1;
  
  if (edge[list[0][2]][list[0][0]] == list[0][1])
    orient[list[0][2]][list[0][0]] = +1;
  else
    orient[list[0][0]][list[0][2]] = -1;
  
  ofound[0] = TRUE;
  nfound++;
  
  more_found = TRUE;
  while (nfound < nf && more_found) {
    more_found = FALSE;
    for (iface = 1; iface<nf; iface++) 
      if (!ofound[iface]) {
	v0 = list[iface][0];
	v1 = list[iface][1];
	v2 = list[iface][2];
	
	if (edge[v0][v1] != v2)
	  face_orient[0] = -orient[v0][v1];
	else
	  face_orient[0] = orient[v1][v0];
	
	if (edge[v1][v2] != v0)
	  face_orient[1] = -orient[v1][v2];
	else
	  face_orient[1] = orient[v2][v1];
	
	if (edge[v2][v0] != v1)
	  face_orient[2] = -orient[v2][v0];
	else
	  face_orient[2] = orient[v0][v2];
	
	down = (face_orient[0] == -1 ||
		face_orient[1] == -1 ||
		face_orient[2] == -1);
	up = (face_orient[0] == +1 ||
	      face_orient[1] == +1 ||
	      face_orient[2] == +1);
	
	if (up && down)
	  return FALSE;
	
	if (up || down) {
	  more_found = TRUE;
	  ofound[iface] = TRUE;
	  nfound++;
	  if (down)
	    new_orient = -1;
	  else
	    new_orient = +1;
	  
	  if (edge[v0][v1] == v2)
	    orient[v0][v1] = new_orient;
	  else
	    orient[v1][v0] = -new_orient;
	  
	  if (edge[v1][v2] == v0)
	    orient[v1][v2] = new_orient;
	  else
	    orient[v2][v1] = -new_orient;
	  
	  if (edge[v2][v0] == v1)
	    orient[v2][v0] = new_orient;
	  else
	    orient[v0][v2] = -new_orient;
	}
      }
  } 
  return TRUE;
}

int admissable(int v0, int v1, int v2)
{
  /* check if adding face (v0,v1,v2) would cause the link of v0 to be a cycle 
     and at least one other edge */
  
  int v3,v4,v5,v6;

  if (fdegree[v0] < 2)
    return TRUE;

  if (equivelar > 0) {
    int new_degree;
    
    if (degree[v0] + 2 > equivelar) {
      new_degree = degree[v0];
      if (edge[v0][v1] == -1 && edge[v1][v0] == -1)
	new_degree++;
      if (edge[v0][v2] == -1 && edge[v2][v0] == -1)
	new_degree++;
      if (new_degree > equivelar)
	return FALSE;
    }
  }    
  
  
  if (degree[v0] == fdegree[v0]+1)
    return TRUE;
  
  if (edge[v0][v1] == -1) {
    if (edge[v1][v0] == -1)
      return TRUE;
    else
      v3 = edge[v1][v0];
  }
  else
    v3 = edge[v0][v1];
  
  if (edge[v0][v2] == -1) {
    if (edge[v2][v0] == -1)
      return TRUE;
    else
      v4 = edge[v2][v0];
  }
  else
    v4 = edge[v0][v2];
  
  v5 = v1;
  while (v3 != v4 && v3 != -1) {
    if (edge[v0][v3] == v5)
      v6 = edge[v3][v0];
    else
      v6 = edge[v0][v3];
    v5 = v3;
    v3 = v6;
  }

  if (v3 == v4)
    return FALSE;
  else
    return TRUE;
}

int smaller_lex()
{
  /* check if the relabeling starting at nt_match is lexigraphically 
     smaller than the current labeling */
  
  int label[3];         /* vertices of face. also labels in relabeling */
  int w0,w1,w2;         /* vertices of face relabeled 
			   (label[0],label[1],label[2]) */
  int i;                /* vertex index for face */
  int branch;           /* search might branch */
  int save_next_label;
  int save_nt_match;

  branch = FALSE;

  for (i=0; i<3; i++)
    label[i] = list[nt_match-1][i];
  
  while (nt_match<nf) {
    
    /* label[0], label[1], and label[2] are the vertices of the nt_match-1 or 
       better */
    /* find the next relabeled face */
    
    w2 = -1;
    while (w2 == -1) {
      
      w0 = label_to_v[label[0]];
      w1 = label_to_v[label[1]];
      
      if (edge[w0][w1] != -1 &&
	  edge[w1][w0] != -1) {
	if (v_to_label[edge[w0][w1]] > label[2] &&
	    v_to_label[edge[w1][w0]] > label[2]) {
	  if (v_to_label[edge[w0][w1]] > v_to_label[edge[w1][w0]])
	    w2 = edge[w1][w0];
	  else
	    w2 = edge[w0][w1];
	}
	else if (v_to_label[edge[w0][w1]] > label[2])
	  w2 = edge[w0][w1];
	else if (v_to_label[edge[w1][w0]] > label[2])
	  w2 = edge[w1][w0];
	else if (v_to_label[edge[w0][w1]] == -1) {
	  
	  if (v_to_label[edge[w1][w0]] == -1)
	    
	    /* both 3rd vertices of the faces of (w0,w1) are unlabeled.  
	       the search must branch below */
	    
	    branch = TRUE;
	  
	  w2 = edge[w0][w1];
	  v_to_label[w2] = next_label;
	  label_to_v[next_label] = w2;
	  next_label++;
	}
	else if (v_to_label[edge[w1][w0]] == -1) {
	  w2 = edge[w1][w0];
	  v_to_label[w2] = next_label;
	  label_to_v[next_label] = w2;
	  next_label++;
	}
      }
      else if (edge[w0][w1] != -1) {
	if (v_to_label[edge[w0][w1]] > label[2])
	  w2 = edge[w0][w1];
	else if (v_to_label[edge[w0][w1]] == -1) {
	  w2 = edge[w0][w1];
	  v_to_label[w2] = next_label;
	  label_to_v[next_label] = w2;
	  next_label++;
	}
      }
      else if (edge[w1][w0] != -1) {
	if (v_to_label[edge[w1][w0]] > label[2])
	  w2 = edge[w1][w0];
	else if (v_to_label[edge[w1][w0]] == -1) {
	  w2 = edge[w1][w0];
	  v_to_label[w2] = next_label;
	  label_to_v[next_label] = w2;
	  next_label++;
	}
      }
      
      if (w2 == -1) {
	if (label[0] == list[nt_match][0] && label[1] == list[nt_match][1])
	  
	  /* current labeling has another face with (label[0],label[1]) but
	     relabeling does not. */
	  
	  return FALSE;
	else {
	  
	  /* neither current labeling nor relabeling have another face with 
	     (label[0],label[1]). */
	  
	  label[1]++;
	  if (label[1] == next_label || label[1] >= nv-1) {
	    label[0]++;
	    label[1] = label[0] + 1;
	  }
	  label[2] = label[1];
	}
      }
      else {
	label[2] = list[nt_match][2];
	if (label[0] == list[nt_match][0] && label[1] == list[nt_match][1]) {
	  if (v_to_label[w2] > label[2])
	    
	    /* current labeling and relabeling both have another face with 
	       (label[0],label[1]).  current labeling is smaller. */
	    
	    return FALSE;
	  else if (v_to_label[w2] < label[2])
	    
	    /* current labeling and relabeling both have another face with 
	       (label[0],label[1]).  relabeling is smaller. */
	    
	    return TRUE;
	  
	  /* current labeling and relabeling both have another face with 
	     (label[0],label[1]).  they are the same. */
	  
	  if (branch) {
	    save_nt_match = nt_match;
	    nt_match++;
	    save_next_label = next_label-1;
	    if (smaller_lex())
	      return TRUE;
	    for ( ; next_label>save_next_label; next_label--)
	      v_to_label[label_to_v[next_label-1]] = -1;
	    nt_match = save_nt_match;
	    branch = FALSE;
	    v_to_label[w2] = -1;
	    w2 = edge[w1][w0];
	    v_to_label[w2] = next_label;
	    label_to_v[next_label] = w2;
	    next_label++;
	  }

	  /* drop out of while loop */
	  
	}
	else
	  
	  /* current labeling does not have another face with 
	     (label[0],label[1]) but relabeling does. */
	  
	  return TRUE;
      }
      
    }
    nt_match++;
  }
  
  /* they are the same */
  
  return FALSE;
}

int min_lex()
{
  /* check that the current labeling of the list of faces is lexigraphical
     minimum */

  int v0,v1,v;
  int ineighbor,ioffset,save;
  int neighbor[MAXN];
  int save_next_label;
  int imatch;

  save_next_label = degree[0]+1;

  for (v0=0; v0<nv; v0++) {
    if (completeness[v0] == 1)
      nmatch[v0] = 0;

    if (completeness[0] != 2 || completeness[v0] != 2) {
      if (degree[v0] == degree[0] &&  complete[v0]) {
	
	/* make a list of the neighbors of v0 */
	
	for (v1=0; edge[v0][v1] == -1; v1++) {}
	neighbor[0] = v1;
	neighbor[1] = edge[v0][v1];
	for (ineighbor=2; ineighbor<degree[0]; ineighbor++)
	  if (edge[v0][neighbor[ineighbor-1]] == neighbor[ineighbor-2])
	    neighbor[ineighbor] = edge[neighbor[ineighbor-1]][v0];
	  else
	    neighbor[ineighbor] = edge[v0][neighbor[ineighbor-1]];
	
	/* set the initial values of v_to_label for v0 */
	
	v_to_label[v0] = 0;
	label_to_v[0] = v0;
	v_to_label[v1] = 1;
	label_to_v[1] = v1;
	for (ioffset=1; ioffset<=degree[0]/2; ioffset++) {
	  v_to_label[neighbor[ioffset]] = 2*ioffset;
	  label_to_v[2*ioffset] = neighbor[ioffset];
	}
	for (ioffset=1; ioffset<(degree[0]+1)/2; ioffset++) {
	  v_to_label[neighbor[-ioffset+degree[0]]] = 2*ioffset+1;
	  label_to_v[2*ioffset+1] = neighbor[-ioffset+degree[0]];
	}
	for (ineighbor=0; ineighbor<degree[0]; ineighbor++) {
	  
	  /* check starting at each neighbor of v0 going clockwise */
	  
	  nt_match = fdegree[v0];
	  next_label = save_next_label;
	  if (smaller_lex()) {
	    for ( ; next_label>0; next_label--)
	      v_to_label[label_to_v[next_label-1]] = -1;
	    return FALSE;
	  }
	  
	  for ( ; next_label>save_next_label; next_label--)
	    v_to_label[label_to_v[next_label-1]] = -1;
	  if (completeness[v0] == 1)
	    if (nt_match >= degree[0]) {
	      for (v=0; v<=degree[0]; v++)
		match[v0][nmatch[v0]][v] = label_to_v[v];
	      nmatch[v0]++;
	    }
	  
	  
	  /* rotate the labeling for next neighbor */
	  
	  save = v_to_label[neighbor[degree[0]-1]];
	  for (ioffset=degree[0]-1; ioffset>0; ioffset--) {
	    v_to_label[neighbor[ioffset]] = v_to_label[neighbor[ioffset-1]];
	    label_to_v[v_to_label[neighbor[ioffset]]] = neighbor[ioffset];
	  }
	  v_to_label[neighbor[0]] = save;
	  label_to_v[v_to_label[neighbor[0]]] = neighbor[0];
	}
	
	/* flip the labeling for counterclockwise check */
	
	for (ioffset=1; ioffset<(degree[0]+1)/2; ioffset++) {
	  save = v_to_label[neighbor[ioffset]];
	  v_to_label[neighbor[ioffset]] = 
	    v_to_label[neighbor[-ioffset+degree[0]]];
	  label_to_v[v_to_label[neighbor[ioffset]]] = neighbor[ioffset];
	  v_to_label[neighbor[-ioffset+degree[0]]] = save;
	  label_to_v[v_to_label[neighbor[-ioffset+degree[0]]]] = 
	    neighbor[-ioffset+degree[0]];
	}
	
	for (ineighbor=0; ineighbor<degree[0]; ineighbor++) {
	  
	  /* check starting at each neighbor of v0 going counterclockwise */
	  
	  nt_match = fdegree[0];
	  next_label = save_next_label;
	  if (smaller_lex()) {
	    for ( ; next_label>0; next_label--)
	      v_to_label[label_to_v[next_label-1]] = -1;
	    return FALSE;
	  }
	  
	  for ( ; next_label>save_next_label; next_label--)
	    v_to_label[label_to_v[next_label-1]] = -1;
	  if (completeness[v0] == 1)
	    if (nt_match >= degree[0]) {
	      for (v=0; v<=degree[0]; v++)
		match[v0][nmatch[v0]][v] = label_to_v[v];
	      nmatch[v0]++;
	    }
	  
	  /* rotate the labeling for next neighbor */
	  
	  save = v_to_label[neighbor[degree[0]-1]];
	  for (ioffset=degree[0]-1; ioffset>0; ioffset--) {
	    v_to_label[neighbor[ioffset]] = v_to_label[neighbor[ioffset-1]];
	    label_to_v[v_to_label[neighbor[ioffset]]] = neighbor[ioffset];
	  }
	  v_to_label[neighbor[0]] = save;
	  label_to_v[v_to_label[neighbor[0]]] = neighbor[0];
	}
	
	/* undo labeling before next v0 or returning */
	
	for ( ; next_label>0; next_label--)
	  v_to_label[label_to_v[next_label-1]] = -1;
      }
    }
    
    else if (nmatch[v0] != 0) {
      
      /* completeness[0] == 2 && completeness[v0] == 2 
	 and lk(0) is equivalent to lk(v0) */
      
      for (imatch=0; imatch<nmatch[v0]; imatch++) {
	for (v=0; v<=degree[0]; v++) {
	  label_to_v[v] = match[v0][imatch][v];
	  v_to_label[label_to_v[v]] = v;
	}
	nt_match = fdegree[0];
	next_label = degree[0] + 1;
	if (smaller_lex()) {
	  for ( ; next_label>0; next_label--)
	    v_to_label[label_to_v[next_label-1]] = -1;
	  return FALSE;
	}
	for ( ; next_label>degree[0]+1; next_label--)
	  v_to_label[label_to_v[next_label-1]] = -1;
      }
      for ( ; next_label>0; next_label--)
	v_to_label[label_to_v[next_label-1]] = -1;
    }
    if (complete[0] && completeness[v0] == 1)
      completeness[v0] = 2;
  }
  
  return TRUE;
}

int contractible(int v0, int v1)
{
  /* check to see if the edge (v0,v1) is contractible.  vertices v0 and v1 are
     complete. */

  int v;
  int neighbors;  /* number of common neighbors of v0 and v1 */

  neighbors = 0;
  for (v=0; v<nv; v++)
    if (edge[v0][v] != -1 && edge[v1][v] != -1) {
      neighbors++;
      if (neighbors == 3)
	return FALSE;
    }
  return TRUE;
}

void found_one(int g, int orient)
{
  /* found a surface */

  char outfilename[100];

  ADDBIG(ngenerated[g][orient],1);
  
  if (FALSE) {
    if (outfile_alpha[nv][g][orient] == NULL) {
      sprintf(outfilename,"2-manifolds_v%d_o%d_g%d_r%s_m%d.alpha%c",
	      nv,orient,g,res_text,mod,0);
      if ((outfile_alpha[nv][g][orient] = fopen(outfilename,"w")) == NULL)
	{
	  fprintf(stderr,
		  "can't open %s for writing\n",outfilename);
	  exit(1);
	}
    }
    
    write_alpha(outfile_alpha[nv][g][orient]);
  }
 
  if (TRUE) {
    if (outfile_lex[nv][g][orient] == NULL) {
      if (equivelar > 0) {
	if (orient)
	  sprintf(outfilename,"equi_surfaces_lex_n%d_p3_q%d_o%d_g%d_r%s_m%d.lex%c",
		  nv,equivelar,orient,g/2,res_text,mod,0);
	else
	  sprintf(outfilename,"equi_surfaces_lex_n%d_p3_q%d_o%d_g%d_r%s_m%d.lex%c",
		  nv,equivelar,orient,g,res_text,mod,0);
      }
      else {
	sprintf(outfilename,"2-manifolds_v%d_o%d_g%d_r%s_m%d.lex%c",
		nv,orient,g,res_text,mod,0);
      }
      if ((outfile_lex[nv][g][orient] = fopen(outfilename,"w")) == NULL)
	{
	  fprintf(stderr,
		  "can't open %s for writing\n",outfilename);
	  exit(1);
	}
    }
    
    write_lex(outfile_lex[nv][g][orient]);
    fflush(outfile_lex[nv][g][orient]);
  }

}

void do_face(int v0, int v1, int v2)
{
  int done;
  int v,g,orient;
  int nattached;

  /* add a new face and check pruning conditions. */

  /* pruning before adding face */

  if (v2 == maxnv)
    return;

  if (nf == maxnf)
    return;

  /* check that adding face would not cause the link of v0, v1, or v2 to be a 
     cycle and at least one other edge */

  if (!admissable(v0,v1,v2))
    return;
  if (!admissable(v1,v2,v0))
    return;
  if (!admissable(v2,v0,v1))
    return;

  /* add a new face */

  add_face(v0,v1,v2);

  /* check that vertex 0 still has minimal degree */
  
  if ((degree[v0] < degree[0] && complete[v0]) ||
      (degree[v1] < degree[0] && complete[v1]) ||
      (degree[v2] < degree[0] && complete[v2]) ) {
    remove_face();
    return;
  }

  /* check for correct orientability */

  if (orientability == 1) {

    /* don't bother if new face is attached by no or one edge or
       if attached by two edges and completes the common vertex */

    nattached = -3;
    if (edge[v0][v1] != -1) nattached++;
    if (edge[v0][v2] != -1) nattached++;
    if (edge[v1][v0] != -1) nattached++;
    if (edge[v1][v2] != -1) nattached++;
    if (edge[v2][v0] != -1) nattached++;
    if (edge[v2][v1] != -1) nattached++;
    if (nattached > 1) {
      if (nattached == 2) {
	if (complete[v0])
	  nattached = 1;
	else if (complete[v1])
	  nattached = 1;
	else if (complete[v2])
	  nattached = 1;
      }
      if (nattached > 1)
	if (!orientable()) {
	  remove_face();
	  return;
	}
    }
  }

  /* pruning after adding face */
  /* only upon completing a vertex */

  if (complete[v0]) 
    if (!min_lex()) {
      remove_face();
      return;
    }
  
  /* check for contractible edge */

  if (only_irreducible && nv > 4) {
    if (complete[v0])
      for (v=0; v<nv; v++)
	if (complete[v] &&
	    edge[v0][v] != -1)
	  if (contractible(v0,v)) {
	    remove_face();
	    return;
	  }
    if (complete[v1])
      for (v=0; v<nv; v++)
	if (complete[v] &&
	    edge[v1][v] != -1 &&
	    v != v0)
	  if (contractible(v1,v)) {
	    remove_face();
	    return;
	  }
    if (complete[v2])
      for (v=0; v<nv; v++)
	if (complete[v] &&
	    edge[v2][v] != -1 &&
	    v != v0 && v != v1)
	  if (contractible(v2,v)) {
	    remove_face();
	    return;
	  }
  }
  
  /* check surface is done */
  
  done = TRUE;
  for (v=v0; done && v<nv; v++)
    done = (complete[v]);

  if (done) {
    if (nf < minnf) {
      remove_face();
      return;
    }

    g = 2-nv+ne-nf;
    if (g < ming || g > maxg) {
      remove_face();
      return;
    }

    /* if orientability == 1 this has already been checked above */

    if (orientability == 1) 
      orient = TRUE;
    else {
      orient = orientable();
      if (orientability == 0 && orient) {
	remove_face();
	return;
      }
    }
    
    if (nv >= minnv && nv <= maxnv) {
      
      /*
	debug_count++;
	if (debug_count == 149)
	dump_it();
      */
      
      found_one(g,orient);
    }
    
    /*
      dump_it();
    */
    
    remove_face();
    return;
  }

  /* do it again */

  next_face();

  remove_face();
}

void next_face() 
{
  /* a face has just been added.  find what the next face might be. */

  int v0,v1,v2;

  if (nf == splitlevel) {
#ifdef SPLITTEST
    splitcases++;
    return;
#endif
    if (splitcount-- != 0) return;
    splitcount = mod - 1;
  }

  v0 = list[nf-1][0];
  v1 = list[nf-1][1];
  v2 = list[nf-1][2]+1;

  /* check if v0 is complete */

  if (complete[v0]) {
    while (complete[v0])
      v0++;
    v1 = v0+1;
    v2 = v1+1;
  }

  /* while edge (v0,v1) is not on boundry */

  while ((edge[v0][v1] == -1 && edge[v1][v0] == -1) ||
	 (edge[v0][v1] != -1 && edge[v1][v0] != -1)) {
    if (edge[v0][v1] == -1 && 
	edge[v1][v0] == -1 &&
	!complete[v1])
      
      /* first use of the edge (v0,v1) */

      for (v2=v1+1; v2<MIN(nv+1,maxnv-1); v2++)
	if ((edge[v0][v2] == -1 || edge[v2][v0] == -1) &&
	    (edge[v1][v2] == -1 || edge[v2][v1] == -1) &&
	    (v2 == nv || !complete[v2]))
	  do_face(v0,v1,v2);
    v1++;
    v2 = v1+1;
  }

  /* edge (v0,v1) is on boundry.  must be used */
   
  for (; v2<nv+1; v2++)
    if ((edge[v0][v2] == -1 || edge[v2][v0] == -1) &&
	(edge[v1][v2] == -1 || edge[v2][v1] == -1) &&
	(v2 == nv || !complete[v2]))
      do_face(v0,v1,v2);
  
}
 
static void
initialize_splitting(int minlevel, int hint, int maxlevel)
  
/* Set splitlevel and splitcount.  */
{
  splitlevel = hint;
  if (splitlevel > maxlevel) splitlevel = maxlevel;
  
  if (splitlevel < minlevel && splitlevel > 0)
    {
      if (minlevel <= maxlevel) splitlevel = minlevel;
      else                      splitlevel = 0;
    }
  if (mod == 1) splitlevel = 0;
  splitcount = res;
}

static int 
getargvalue(char *arg)
  
/* Find integer value of arguement. 
   arg is a pointer to a command-line argument.
   The value of the arguement is the function return value.
   An absent value is equivalent to 0.
*/
{
  int j,ans,neg;
  
  ans = 0;
  if (arg[0] == '-') {
    neg = -1;
    j = 1;
  }
  else {
    neg = +1;
    j = 0;
  }
  
  for (; arg[j] >= '0' && arg[j] <= '9'; ++j)
    ans = ans * 10 + (arg[j] - '0');
  
  if (arg[j] != '\0')
    return -999999;
  else
    return neg*ans;
}

void summarize() 
{
  /* summarize results */

  int g;

  if (verbose) {
    for (g=0; g<MAXG; g++)
      if (!ISZEROBIG(ngenerated[g][1])) {
	PRINTBIG(stderr,ngenerated[g][1]);
	fprintf(stderr," generated, Euler genus %d, orientable\n",g);
      }
    for (g=0; g<MAXG; g++)
      if (!ISZEROBIG(ngenerated[g][0])) {
	PRINTBIG(stderr,ngenerated[g][0]);
	fprintf(stderr," generated, Euler genus %d, nonorientable\n",g);
      }
  }
  
}

int
main(argc,argv)
int argc;
char *argv[];

{
  int iarg;
  int badargs;
  char *arg;
  int hint;
  int degree0;
  int lg,lnv,lmindegree;
  int res_digits;

#if CPUTIME
  struct tms timestruct0,timestruct1;
  
  times(&timestruct0);
#endif
 
  only_irreducible = FALSE;
  orientability = -1;
  equivelar = -1;
  mindegree = -1;
  minnv = 4;
  maxnv = MAXN;
  minnf = 4;
  maxnf = MAXF;
  ming = 0;
  maxg = MAXG;
  res = 0;
  mod = 1;
  badargs = FALSE;

  for (iarg = 0; iarg < argc ; ++iarg) fprintf(stderr,"%s ",argv[iarg]);
  fprintf(stderr,"\n");

  if (argc < 2) {
    fprintf(stderr,"%s\n",HELPTEXT);
    exit(1);
  }

  iarg = 1;
  while (iarg < argc) {
    arg = argv[iarg];
    if (strcmp(arg,"-v") == 0) {
      verbose = TRUE;
      iarg++;
    }
    else if (strcmp(arg,"-h") == 0) {
      fprintf(stderr,"%s\n",HELPTEXT);
      exit(1);
    }
    else if (strcmp(arg,"-r") == 0) {
      iarg++;
      if (iarg >= argc) {
	fprintf(stderr,"res value required after -r switch\n");
	badargs = TRUE;
      }
      else {
	res = getargvalue(argv[iarg]);
	iarg++;
	if (res < 0) {
	  fprintf(stderr,"res (%d) must be nonnegative.\n",res);
	  badargs = TRUE;
	  res = 0;
	}
      }
    }
    else if (strcmp(arg,"-m") == 0) {
      iarg++;
      if (iarg >= argc) {
	fprintf(stderr,"mod value required after -m switch\n");
	badargs = TRUE;
      }
      else {
	mod = getargvalue(argv[iarg]);
	iarg++;
	if (mod < 1) {
	  fprintf(stderr,"mod (%d) must be posotive.\n",mod);
	  badargs = TRUE;
	  mod = 1;
	}
      }
    }
    else if (strcmp(arg,"-o") == 0) {
      if (orientability == 0 || orientability == 1) {
	fprintf(stderr,"o switch is redefining orientability\n");
	badargs = TRUE;
      }
      iarg++;
      if (iarg >= argc) {
	fprintf(stderr,"value required after -o switch\n");
	badargs = TRUE;
      }
      else {
	orientability = getargvalue(argv[iarg]);
	iarg++;
	if (orientability != 0 && orientability != 1) {
	  fprintf(stderr,"orientability (%d) must be 0 or 1.\n",
		  orientability);
	  badargs = TRUE;
	  orientability = -1;
	}
      }
    }
    else if (strcmp(arg,"-e") == 0) {
      iarg++;
      if (iarg >= argc) {
	fprintf(stderr,"value required after -e switch\n");
	badargs = TRUE;
      }
      else {
	equivelar = getargvalue(argv[iarg]);
	iarg++;
	if (equivelar < 3) {
	  fprintf(stderr,"equivelar degree must be at least 3\n");
	  badargs = TRUE;
	  equivelar = 3;
	}
      }
    }
    else if (strcmp(arg,"-d") == 0) {
      iarg++;
      if (iarg >= argc) {
	fprintf(stderr,"value required after -d switch\n");
	badargs = TRUE;
      }
      else {
	mindegree = getargvalue(argv[iarg]);
	iarg++;
	if (mindegree < 3) {
	  fprintf(stderr,"minimum degree must be at least 3\n");
	  badargs = TRUE;
	  mindegree = 3;
	}
      }
    }
    else if (strcmp(arg,"-g") == 0) {
      iarg++;
      if (iarg >= argc) {
	fprintf(stderr,"value required after -g switch\n");
	badargs = TRUE;
      }
      else {
	ming = getargvalue(argv[iarg]);
	iarg++;
      }
    }
    else if (strcmp(arg,"-G") == 0) {
      iarg++;
      if (iarg >= argc) {
	fprintf(stderr,"value required after -G switch\n");
	badargs = TRUE;
      }
      else {
	maxg = getargvalue(argv[iarg]);
	iarg++;
      }
    }
    else if (strcmp(arg,"-i") == 0) {
      only_irreducible = TRUE;
      iarg++;
    }
    else if (iarg < argc-1) {
      iarg++;
      fprintf(stderr,"unknown switch %s\n",arg);
      badargs = TRUE;
    }
    else {
      maxnv = getargvalue(arg);
      iarg++;
      if (maxnv == -999999) {
	fprintf(stderr,"unknown switch %s\n",arg);
	badargs = TRUE;
      }
    }
  }

  if (equivelar < 3)
    minnv = maxnv;

  if (maxnv > MAXN) {
    fprintf(stderr,"maxnv (%d) too large.  increase MAXN.\n",maxnv);
    badargs = TRUE;
  }

  if (minnv < 4) {
    fprintf(stderr,"minnv (%d) too small.\n",minnv);
    badargs = TRUE;
  }

  if (minnv > maxnv) {
    fprintf(stderr,"minnv (%d) > maxnv (%d).\n",minnv,maxnv);
    badargs = TRUE;
  }

  if (maxg > MAXG) {
    fprintf(stderr,"maxg (%d) too large.  increase MAXG.\n",maxg);
    badargs = TRUE;
  }

  if (ming < 0) {
    fprintf(stderr,"ming (%d) too small.\n",ming);
    badargs = TRUE;
  }

  if (ming > maxg) {
    fprintf(stderr,"ming (%d) > maxg (%d).\n",ming,maxg);
    badargs = TRUE;
  }

  if (maxg > (maxnv-4)*(maxnv-3)/6) {
    fprintf(stderr,"maxg (%d) is not obtainable for specified maxnv (%d).\n",
	    maxg,maxnv);
    maxg = (maxnv-4)*(maxnv-3)/6;
    fprintf(stderr,"reducing maxg to %d.\n",maxg);
    if (ming > maxg) {
      fprintf(stderr,"ming (%d) > maxg (%d).\n",ming,maxg);
      badargs = TRUE;
    }
  }

  if (orientability == 1 && maxg%2 != 0) {
    fprintf(stderr,"Only orientable triangulations and maxg (%d) is odd.\n",
	    maxg);
    maxg--;
    fprintf(stderr,"reducing maxg to %d.\n",maxg);
    if (ming > maxg) {
      fprintf(stderr,"ming (%d) > maxg (%d).\n",ming,maxg);
      badargs = TRUE;
    }
  }

  if (orientability == 1 && ming%2 != 0) {
    fprintf(stderr,"Only orientable triangulations and ming (%d) is odd.\n",
	    ming);
    ming++;
    fprintf(stderr,"increasing ming to %d.\n",ming);
    if (ming > maxg) {
      fprintf(stderr,"ming (%d) > maxg (%d).\n",ming,maxg);
      badargs = TRUE;
    }
  }

  if (ming > (minnv-4)*(minnv-3)/6) {
    fprintf(stderr,"ming (%d) is not obtainable for specified minnv (%d).\n",
	    ming,minnv);
    while (ming > (minnv-4)*(minnv-3)/6)
      minnv++;
    fprintf(stderr,"increasing minnv to %d.\n",minnv);
    if (minnv > maxnv) {
      fprintf(stderr,"minnv (%d) > maxnv (%d).\n",minnv,maxnv);
      badargs = TRUE;
    }
  }
  
  minnf = 2*(ming-2+minnv);
  maxnf = 2*(maxg-2+maxnv);

  if (maxnf >= MAXF) {
    fprintf(stderr,"number of faces (%d) too large.  increase MAXF.\n",
	    maxnf);
    badargs = TRUE;
  }

  maxdegree = maxnv-1;
  if (mindegree == -1) {
    mindegree = maxnv;
    
    for (lg=ming; lg<=maxg; lg++)
      for (lnv=minnv; lnv<=maxnv; lnv++) {
	lmindegree = MAX(3,6*(lg-2+lnv)-(lnv-1)*(lnv-1));
	if (lmindegree < mindegree)
	  mindegree = lmindegree;
      }
  }

  if (equivelar > 0) {
    if (mindegree > equivelar || maxdegree < equivelar) {
      fprintf(stderr,
	      "equivelar (%d) not between mindegree (%d) and maxdegree (%d).\n",
	      equivelar,mindegree,maxdegree);
      badargs = TRUE;
    }
    else {
      mindegree = equivelar;
      maxdegree = equivelar;
    }
  }

  if (res >= mod) {
    fprintf(stderr,"res (%d) >= mod (%d).\n",res,mod);
    badargs = TRUE;
    res = 0;
    mod = 1;
  }

  if (badargs) {
    fprintf(stderr,"Usage: %s\n",USAGE);
    exit(1);
  }
  
#ifdef SPLITTEST
  if (mod == 1) mod = 2;
#endif

  if (verbose) {
    fprintf(stderr,"Minimum number of vertices: %d\n", minnv);
    fprintf(stderr,"Maximum number of vertices: %d\n", maxnv);
    fprintf(stderr,"Minimum number of faces: %d\n", minnf);
    fprintf(stderr,"Maximum number of faces: %d\n", maxnf);
    fprintf(stderr,"Minimum Euler genus: %d\n", ming);
    fprintf(stderr,"Maximum Euler genus: %d\n", maxg);
    fprintf(stderr,"Minimum degree of vertices: %d\n", mindegree);
    fprintf(stderr,"Maximum degree of vertices: %d\n", maxdegree);
    if (orientability == 0)
      fprintf(stderr,"Only nonorientable are generated\n");
    else if (orientability == 1)
      fprintf(stderr,"Only orientable are generated\n");
    else
      fprintf(stderr,"Either nonorientable or orientable are generated\n");
    if (only_irreducible)
      fprintf(stderr,"Only irreducible triangulations are generated.\n");
    fprintf(stderr,"res/mod = %d/%d\n",res,mod);
  }

  hint = MIN(maxnf/2, (3*maxnv)/2);
  if (maxnv <= 16)
    hint = MIN(hint, maxnv+7);
  else if (maxnv <= 19)
    hint = MIN(hint, 24);
  else 
    hint = MIN(hint, maxnv+4);

  initialize_splitting(4,hint,maxnf-1);

  if (verbose && mod != 1)
    fprintf(stderr,"splitlevel is at %d faces.\n",splitlevel);

  sprintf(res_text,"%d",mod-1);
  res_digits = strlen(res_text);
  sprintf(res_text,"%c%d.%dd",'%',res_digits,res_digits);
  sprintf(res_text,res_text,res);

  initialize();

  for (degree0=mindegree; degree0 <= maxdegree; degree0++) {
    
    initialize_first_vertex(degree0);
    
    next_face();
  }

#if CPUTIME
  times(&timestruct1);
#endif
  
#ifdef SPLITTEST
  fprintf(stderr,"%d splitting cases at level = %d",splitcases,splitlevel);
#if CPUTIME
  fprintf(stderr,"; cpu=%.2f sec\n",
	  (double)(timestruct1.tms_utime+timestruct1.tms_stime
		   -timestruct0.tms_utime-timestruct0.tms_stime) / 
	  (double)CLK_TCK);
#else
  fprintf(stderr,"\n");
#endif
  return 0;
#endif
  
  summarize();
  
#if CPUTIME
  fprintf(stderr,"cpu=%.2f sec\n",
	  (double)(timestruct1.tms_utime+timestruct1.tms_stime
		   -timestruct0.tms_utime+timestruct0.tms_stime) / 
	  (double)CLK_TCK);
#else
  fprintf(stderr,"\n");
#endif

  exit(0);
}
